/* MakeConverter.m created by cwang on Thu 06-Apr-2000 

   Openstep2GNU converter is used to convert Openstep makefiles to GNUstep
   makefiles called GNUmakefile files.

   Run Openstep2GNU without parameters to convert makefiles in current directory.

   Run Openstep2GNU -R in the top level directory to recursively convert
   all Openstep makefiles to GNUstep makefiles.


   For CCS users:
   -------------
   To build the GNUmakefile after the conversion, make sure you check out
   the CCS modified GNUstep bundle.make from devman under COMMON/GNUSTEP and
   copied it into the GNUstep/Makefile directory where you have GNUstep installed.

*/

#import "MakeConverter.h"

#define APP_INFO_FILE		@"appDefine.info"
#define BUNDLE_INFO_FILE	@"bundleDefine.info"
#define PB_PROJECT			@"PB.Project"
#define PB_PROJECT2			@"PB.project"
#define MAKE				@"Makefile"
#define MAKE_PREAMBLE		@"Makefile.preamble"
#define MAKE_POSTAMBLE		@"Makefile.postamble"
#define GNU_MAKE			@"GNUmakefile"
#define GNU_MAKE_PREAMBLE	@"GNUmakefile.preamble"
#define GNU_MAKE_POSTAMBLE	@"GNUmakefile.postamble"

#define PROJECT_NAME		@"NAME"
#define PROJECT_TYPE		@"PROJECT_TYPE"
#define PRINCIPAL_CLASS		@"APPCLASS"
#define M_FILES				@"MFILES"

NSDictionary	*gKeysDictionary	= nil;
NSDictionary	*gTypesDictionary	= nil;
NSDictionary	*gMakeDictionary	= nil;

@implementation MakeConverter

/*
    gKeysDictionary
 
    Used in translation of Openstep Makefile defines to GNU Makefile defines
*/
+ gKeysDictionary
{
    if( !gKeysDictionary )
        gKeysDictionary = [NSDictionary dictionaryWithObjectsAndKeys:
            @"OBJC_FILES",				@"CLASSES",
            @"OBJC_FILES",				@"MFILES",
            @"HEADER_FILES",			@"HFILES",
            @"RESOURCE_FILES",			@"GLOBAL_RESOURCES",
            @"PRINCIPAL_CLASS",			@"APPCLASS",
            @"ADDITIONAL_INSTALL_DIRS",	@"CCS_CUSTOM_CLIENT",
            @"ADDITIONAL_INCLUDE_DIRS",	@"CCS_CUSTOM_CLIENT_HEADER_DIRS",
            nil];
    return gKeysDictionary;
}
- keysDictionary	{	return [[self class] gKeysDictionary];	}

/*
    gTypesDictionary

    Used in translation of Openstep project type defines to GNU project type defines
*/
+ gTypesDictionary
{
    if( !gTypesDictionary )
        gTypesDictionary = [NSDictionary dictionaryWithObjectsAndKeys:
            @"BUNDLE_NAME",		@"Loadable Bundle",
            @"BUNDLE_NAME",		@"Palette",
            @"APP_NAME",		@"Application",
            @"TOOL_NAME",		@"Tool",
            @"LIBRARY_NAME",	@"Library",
            nil];
    return gTypesDictionary;
}
- gTypesDictionary	{	return [[self class] gTypesDictionary];	}

/*
    gMakeDictionary

    Determines which make build we should use to compile the GNU make file
*/
+ gMakeDictionary
{
    if( !gMakeDictionary )
        gMakeDictionary = [NSDictionary dictionaryWithObjectsAndKeys:
            @"include $(GNUSTEP_MAKEFILES)/bundle.make",		@"Loadable Bundle",
            @"include $(GNUSTEP_MAKEFILES)/bundle.make",		@"Palette",
            @"include $(GNUSTEP_MAKEFILES)/application.make",	@"Application",
            @"include $(GNUSTEP_MAKEFILES)/tool.make",			@"Tool",
            @"include $(GNUSTEP_MAKEFILES)/library.make",		@"Library",
            nil];
    return gMakeDictionary;
}
- makeDictionary	{	return [[self class] gMakeDictionary];	}

- initWithDir:aDir
{
    if( !aDir )
        return nil;

    [super init];

    fileManager = [NSFileManager defaultManager];
    if( aDir )
        ourDir = [aDir copy];
    return self;
}

+ converterForDirectory:(NSString *)aDir
{
    return [[[self alloc] initWithDir:aDir] autorelease];
}

/*
    gnuTypeForPBType

    Determines the project type from Openstep Makefile
*/
- (NSString *)gnuTypeForPBType:(NSString *)pbType
{
    NSString	*result	= nil;
    result = [[self gTypesDictionary] objectForKey:pbType];
    if( result )
        return result;
    else
        return [@"Unknown_project_type_in_conversion_" stringByAppendingString:pbType];
}

/*
    parseKeyFromLine

    Retrieves the Openstep Makefile defines
*/
- (NSString *)parseKeyFromLine:(NSString *)line
{
    NSString	*result	= nil;
    NSRange	equalRange	= [line rangeOfString:@" ="];

    if( equalRange.length )
        result = [line substringToIndex:equalRange.location];

    return result;
}

/*
    parseKeyFromLine

    Tries to retrieve Openstep Makefile "-include" or "include" from the line
*/
- (NSString *)checkForIncludeFromLine:(NSString *)line
{
    NSString	*result	= nil;
    NSRange	equalRange	= [line rangeOfString:@" "];

    if( equalRange.length )
        result = [line substringToIndex:equalRange.location];

    return result;
}

/*
    parseValueFromLine

    Retrieves the values from the define keys
*/
- (NSString *)parseValueFromLine:(NSString *)line
{
    NSString	*result	= nil;
    NSRange	equalRange	= [line rangeOfString:@"= "];

    if( equalRange.length )
    {
        int		start		= equalRange.location + equalRange.length;

        if( start < [line length] )
        {
            NSString *temp = [line substringFromIndex:start];

            if( [temp length] )
            {
                NSScanner *scanner	= [NSScanner scannerWithString:temp];
                NSCharacterSet	*whiteSet			= [NSCharacterSet whitespaceCharacterSet];
                // need the \r to remove the naughty ^M's or things won't work
                NSCharacterSet	*newline	= [NSCharacterSet characterSetWithCharactersInString:@"\n\r"];

                [scanner setCharactersToBeSkipped:whiteSet];
                [scanner scanUpToCharactersFromSet:newline
                                        intoString:&result];
            }
        }
    }

    return result;
}

/*
    fixBackSlash

    Fixes the \ from Openstep Makefile to the GNUmakefile
*/
- (NSMutableString *)fixBackSlash:(NSString *)string
{
    //NSMutableString	*resultString	= [NSMutableString stringWithString:string];
    NSMutableString	*resultString	= [NSMutableString string];
    NSRange 		range 			= [string rangeOfString:@"\\" options:NSBackwardsSearch];
/*
    if ( range.location != NSNotFound )
    {
        [resultString replaceCharactersInRange:range withString:@" \\"];
    }
*/
    if ( range.location != NSNotFound )
    {
        [resultString appendString:[string substringToIndex:range.location]];
        [resultString appendString:@" \\"];
    }
    else
        [resultString appendString:string];

    return resultString;
}

/*
    stripControlMs

    Strip the ^M's from the lines
*/
- (NSMutableString *)stripControlMs:(NSString *)string
{
    //NSMutableString	*resultString	= [[string substringToIndex:[string length] - 1] mutableCopy];
    
    NSMutableString	*resultString	= [NSMutableString string];
    NSMutableString	*newString		= nil;
    NSCharacterSet	*whiteSet		= [NSCharacterSet whitespaceCharacterSet];
    NSMutableCharacterSet *charset	= [[NSMutableCharacterSet controlCharacterSet] mutableCopy];
    id	scanner = [NSScanner scannerWithString:string];

    [charset addCharactersInString:@"\t"];

    [scanner setCharactersToBeSkipped:whiteSet];
    [scanner scanUpToCharactersFromSet:charset intoString:&newString];

    if (newString)
        [resultString appendFormat:@"%@\n", newString];
    else
        [resultString appendFormat:@"\n"];

    return resultString;
}

/*
    convertPBMakefilePreamble

    Here we create the GNUmakefile.preamble
*/
- (NSString *)convertPBMakefilePreamble
{
    int				i;
    NSArray			*lines;
    NSMutableString	*resultString	= [NSMutableString string];
    //BOOL			result			= NO;
    NSDictionary	*keysDictionary	= [self keysDictionary];

    // we get the lines in Openstep Makefile.preamble into an array
    lines = [pbMakefilePreamble componentsSeparatedByString:@"\n"];

    gnuMakefilePreamble = [NSMutableString string];

    // now we process each line accordingly
    for (i = 0; i < [lines count]; i++)
    {
        NSString	*line 		= [self fixBackSlash:[lines objectAtIndex:i]];
        NSString	*newLine	= [line stringByAppendingString:@"\n"];
        NSString	*key		= [self parseKeyFromLine:line];

        // translate the define key for CCS_CUSTOM_CLIENT to ADDITIONAL_INSTALL_DIRS
        if( key && [keysDictionary objectForKey:key] )
        {
            NSString *newKey	= [keysDictionary objectForKey:key];
            NSString *val		= [self parseValueFromLine:line];

            // want to make sure there is a value associated with the define key
            // or else we'll be creating directories with *nil* names when we do
            // a GNU make
            if ( [val length] )
            {
                newLine = [NSString stringWithFormat:@"%@ = %@\n", newKey, val];
            }
        }
        
        if( newLine )
            [resultString appendString:[self stripControlMs:newLine]];
    }

    if( [resultString length] )
    {
        return resultString;
    }
    else
        return nil;
}

/*
    convertPBMakefile

    Here we create the GNUmakefile
*/
- (NSString *)convertPBMakefile
{
    int				i;
    NSArray			*lines;
    NSMutableString	*resultString	= [NSMutableString string];
    BOOL			result			= NO;
    NSString		*projectName	= nil;
    NSString		*projectType	= nil;

    // we get the lines in Openstep Makefile into an array
    lines = [pbMakefile componentsSeparatedByString:@"\n"];

    gnuMakefile = [NSMutableString string];
    pbKeys		= [NSMutableDictionary dictionary];

    // here we setup the key value pairs for determining project type
    for (i = 0; i < [lines count]; i++)
    {
        NSString *line = [lines objectAtIndex:i];
        NSString *aKey	= [self parseKeyFromLine:line];

        if( aKey )
        {
            NSString	*aVal	= [self parseValueFromLine:line];

            if( aVal )
                [pbKeys setObject:aVal forKey:aKey];
        }
    }

    projectName	= [pbKeys objectForKey:PROJECT_NAME];
    projectType	= [pbKeys objectForKey:PROJECT_TYPE];

    // we will do the conversion if there is a project name and type
    if( [projectName length] && [projectType length] )
    {
        NSDictionary	*keysDictionary	= [self keysDictionary];
        NSDictionary	*makeDictionary	= [self makeDictionary];
        result = YES;

        // need this env check so that when u have GNUmakefile in the Windows environment
        // you don't build with GNUmakefile instead of pb Makefile
        [resultString appendString:@"ifeq \"$(GNUSTEP_SYSTEM_ROOT)\" \"\"\n"];
        [resultString appendString:@"  include Makefile\n"];
        [resultString appendString:@"else\n\n"];

        // put in some GNUmakefile stuff
        [resultString appendString:@"GNUSTEP_INSTALLATION_DIR = $(GNUSTEP_SYSTEM_ROOT)\n"];
        [resultString appendString:@"GNUSTEP_MAKEFILES = $(GNUSTEP_SYSTEM_ROOT)/Makefiles\n"];
        [resultString appendString:@"include $(GNUSTEP_MAKEFILES)/common.make\n"];

        // see if we need to add the principal class to the GNUmakefile from
        // Openstep PB.project file
        if ( [[pbDictionary objectForKey:PRINCIPAL_CLASS] length] )
        {
            [resultString appendString:projectName];
            [resultString appendString:@"_PRINCIPAL_CLASS = "];
            [resultString appendString:[pbDictionary objectForKey:PRINCIPAL_CLASS]];
        }
        
        [resultString appendString:@"\n\n"];

        // now we process each line accordingly
        for (i = 0; i < [lines count]; i++)
        {
            NSString	*line 		= [self fixBackSlash:[lines objectAtIndex:i]];
            NSString	*newLine	= [line stringByAppendingString:@"\n"];
            NSString	*checkLine  = [self checkForIncludeFromLine:line];

            // set the project name
            if( [line hasPrefix:PROJECT_NAME] )
            {
                newLine = [NSString stringWithFormat:@"%@ = %@\n",
                    [self gnuTypeForPBType:projectType], projectName];
            }
            else if ( (checkLine && [checkLine hasSuffix:@"include"]) ||
                      [line hasPrefix:PROJECT_NAME] )
            {
                // skip the pb Makefile include statements b/c
                // we're adding the GNUmakefile ones
                continue;
            }
            else
            {
                NSString	*key	= [self parseKeyFromLine:line];
                // translate the define keys
                if( key && [keysDictionary objectForKey:key] )
                {
                    NSString *newKey	= [keysDictionary objectForKey:key];
                    NSString *val		= [self parseValueFromLine:line];

                    if ( [line hasPrefix:M_FILES] )
                    {
                        // want to include this definition in OBJC_FILES
                        newLine = [NSString stringWithFormat:@"%@_%@ += %@\n",
                            projectName, newKey, val];
                    }
                    else
                    {
                        newLine = [NSString stringWithFormat:@"%@_%@ = %@\n",
                            projectName, newKey, val];
                    }
                }
            }

            if( newLine )
                [resultString appendString:[self stripControlMs:newLine]];
        }

        // we include the GNUmakefile includes
        [resultString appendString:@"-include GNUmakefile.preamble\n"];
        [resultString appendString:[self stripControlMs:[makeDictionary objectForKey:projectType]]];
        [resultString appendString:@"\n"];
        [resultString appendString:@"-include GNUmakefile.postamble\n\n"];

        // close the if statement
        [resultString appendString:@"endif\n"];
    }
    else
        NSLog( @"Couldn't parse project name & type" );

    if( result )
    {
        return resultString;
    }
    else
        return nil;
}

/*
    convertDirectoryIfNeeded

    we see if we need to convert the makefiles in this directory
*/
- (void)convertDirectoryIfNeeded
{
    NSString	*pbProjectPath	= [ourDir stringByAppendingPathComponent:PB_PROJECT];

    if( ! [fileManager fileExistsAtPath:pbProjectPath] )
      pbProjectPath	= [ourDir stringByAppendingPathComponent:PB_PROJECT2];

    if( [fileManager fileExistsAtPath:pbProjectPath] )
    {
        NSString	*pbPath	= [ourDir stringByAppendingPathComponent:PB_PROJECT];
        NSString	*pbMakefilePath		= [ourDir stringByAppendingPathComponent:MAKE];
        NSString	*pbPreamblePath		= [ourDir stringByAppendingPathComponent:MAKE_PREAMBLE];
        NSString	*gnuMakefilePath	= [ourDir stringByAppendingPathComponent:GNU_MAKE];
        NSString	*gnuPreamblePath	= [ourDir stringByAppendingPathComponent:GNU_MAKE_PREAMBLE];

        pbDictionary = [[NSDictionary dictionaryWithContentsOfFile:pbPath] retain];
        // we look for some pb files in directory to determine if we need to go on
        if( [fileManager fileExistsAtPath:pbMakefilePath] )
        {
            // we convert the the pb makefile and makefile.preamble
            pbMakefile  = [NSString stringWithContentsOfFile:pbMakefilePath];
            gnuMakefile = [self convertPBMakefile];
            pbMakefilePreamble  = [NSString stringWithContentsOfFile:pbPreamblePath];
            gnuMakefilePreamble = [self convertPBMakefilePreamble];

            // create gnu makefiles
            if ( [gnuMakefile length] )
            {
                NSLog(@"Converted GNUMakefile = %@", gnuMakefilePath);
                [gnuMakefile writeToFile:gnuMakefilePath atomically:YES];
            }
            else
            {
                NSLog(@"convertion failed GNUMakefile = %@", gnuMakefilePath);
            }

            // create gnu makefiles.preamble
            if ( [gnuMakefilePreamble length] )
            {
                NSLog(@"Converted GNUMakefile.preamble = %@", gnuPreamblePath);
                [gnuMakefilePreamble writeToFile:gnuPreamblePath atomically:YES];
            }
            else
            {
                NSLog(@"convertion failed GNUMakefile.preamble = %@", gnuPreamblePath);
            }

        }
    }
}

/*  -------------------------------------------------

           Main  convert makefile controller

   -------------------------------------------------- */
+ (void)convertCurrentDirectory:(BOOL)recurse
{
    id 				dir 			= [[NSFileManager defaultManager] currentDirectoryPath];
    id				aConverter		= [self converterForDirectory:dir];
    // recursively go into the root directory to retrieve file/directory item enumerated paths
    NSDirectoryEnumerator	*directoryEnum	= [[NSFileManager defaultManager] enumeratorAtPath:dir];
    NSString 				*file;
    BOOL					isDir;

    //NSLog(@"current dir= %@", dir);
    //NSLog(@"dir content= %@", pathContents);

    // do makefile convertion in the current directory
    [aConverter convertDirectoryIfNeeded];

    // see if we need to recursively go into directories
    if ( recurse )
    {
        // now process files in the subdirectories if needed
        while ( (file = [directoryEnum nextObject]) )
        {
            //NSLog(@"root dir content= %@", file);
            if( [[NSFileManager defaultManager] fileExistsAtPath:file isDirectory:&isDir] )
            {
                // we do it for directories not files
                if( isDir )
                {
                    //NSLog(@"we are in directory= %@", file);
                    [[self converterForDirectory:file] convertDirectoryIfNeeded];
                }
            }
        }
    }
}

@end
